import logging
from collections import defaultdict
from collections.abc import Callable

from lato.message import Message
from lato.utils import OrderedSet

log = logging.getLogger(__name__)


class ApplicationModule:
    def __init__(self, name: str):
        """Initialize the application module instance.

        :param name: Name of the module
        """
        self.name: str = name
        self._handlers: defaultdict[str, OrderedSet[Callable]] = defaultdict(OrderedSet)
        self._submodules: OrderedSet[ApplicationModule] = OrderedSet()

    def include_submodule(self, a_module: "ApplicationModule"):
        """Adds a child submodule to this module.

        :param a_module: child module to add
        """
        assert isinstance(
            a_module, ApplicationModule
        ), f"Can only include {ApplicationModule} instances, got {a_module}"
        self._submodules.add(a_module)

    def handler(self, alias: type[Message] | str):
        """
        Decorator for registering a handler. Handler can be aliased by a name or by a message type.

        :param alias: :class:`lato.Message` or a string.

        Example #1:
        -----------
        >>> from lato import Application, ApplicationModule
        >>> my_module = ApplicationModule("my_module")
        >>>
        >>> @my_module.handler("my_handler")
        >>> def my_handler():
        >>>     print("handler called")
        >>>
        >>> app = Application("example")
        >>> app.include_submodule(my_module)
        >>> app.call("my_handler")
        "handler called"

        Example #2:
        -----------
        >>> from lato import ApplicationModule, Command
        >>> class MyCommand(Command):
        >>>     ...
        >>>
        >>> my_module = ApplicationModule("my_module")
        >>> @my_module.handler(MyCommand)
        >>> def my_handler(command: MyCommand):
        >>>     print("command handler called")
        >>>
        >>> app = Application("example")
        >>> app.include_submodule(my_module)
        >>> app.execute(MyCommand())
        command handler called

        """
        try:
            is_message = issubclass(alias, Message)
        except TypeError:
            is_message = False

        if callable(alias) and not is_message:
            # decorator was called without any argument
            func = alias
            alias = func.__name__
            assert len(self._handlers[alias]) == 0
            self._handlers[alias].add(func)
            return func

        # decorator was called with argument
        # @my_module.handle("my_function")
        # @my_module.handle(MyCommand)
        def decorator(func):
            """
            Decorator for registering tasks by name
            """
            self._handlers[alias].add(func)
            return func

        return decorator

    def iterate_handlers_for(self, alias: str):
        if alias in self._handlers:
            for handler in self._handlers[alias]:
                yield handler
        for submodule in self._submodules:
            try:
                yield from submodule.iterate_handlers_for(alias)
            except KeyError:
                pass

    def get_handlers_for(self, alias: str):
        return list(self.iterate_handlers_for(alias))

    def __repr__(self):
        return f"<{self.name} {object.__repr__(self)}>"



import logging
from collections.abc import Callable
from typing import Any

from lato.application_module import ApplicationModule
from lato.dependency_provider import BasicDependencyProvider, DependencyProvider
from lato.message import Command, Event, Message
from lato.transaction_context import TransactionContext

log = logging.getLogger(__name__)


class Application(ApplicationModule):
    """Core Application class.

    This class represents the core functionality of an application and extends :class:`ApplicationModule`.

    :param dependency_provider_factory: a factory that returns :class:`DependencyProvider` instance,
        defaults to BasicDependencyProvider.
    """

    dependency_provider_factory: type[DependencyProvider] = BasicDependencyProvider

    def __init__(
        self,
        name=__name__,
        dependency_provider: DependencyProvider | None = None,
        **kwargs,
    ):
        """Initialize the application instance.

        :param name: Name of the application
        :param dependency_provider: dependency provider :class:`DependencyProvider` instance.
                                    Defaults to :class:`BasicDependencyProvider` instance populated with kwargs.
        :param kwargs: Additional keyword arguments to be passed to the dependency provider.

        """
        super().__init__(name)
        self.dependency_provider = (
            dependency_provider or self.dependency_provider_factory(**kwargs)
        )
        self._transaction_context_factory = None
        self._on_enter_transaction_context = lambda ctx: None
        self._on_exit_transaction_context = lambda ctx, exception=None: None
        self._transaction_middlewares = []
        self._composers: dict[str | Command, Callable] = {}

    def get_dependency(self, identifier: str | type) -> Any:
        """Gets a dependency from the dependency provider. Dependencies can be resolved either by name or by type.

        :param identifier: A string or a type representing the dependency.

        :return: The resolved dependency.
        """
        return self.dependency_provider.get_dependency(identifier)

    def __getitem__(self, identifier: str | type) -> Any:
        return self.get_dependency(identifier)

    def call(self, func: Callable | str, *args, **kwargs):
        """Invokes a function with `args` and `kwargs` within the :class:`TransactionContext`.
        If `func` is a string, then it is an alias, and the corresponding handler for the alias is retrieved.
        Any missing arguments are provided by the dependency provider of a transaction context,
        and args and kwargs parameters.

        :param func: The function to invoke, or an alias.
        :param args: Arguments to pass to the function.
        :param kwargs: Keyword arguments to pass to the function.

        :return: The result of the invoked function.

        :raises ValueError: If an alias is provided, but no corresponding handler is found.
        """
        if isinstance(func, str):
            try:
                func = next(self.iterate_handlers_for(alias=func))
            except StopIteration:
                raise ValueError(f"Handler not found", func)

        with self.transaction_context() as ctx:
            result = ctx.call(func, *args, **kwargs)
        return result

    def execute(self, message: Message) -> Any:
        """Executes a command within the :class:`TransactionContext`.
        Use :func:`handler` decorator to register a handler for the command.
        If a command is handled by multiple handlers, then the final result is
        composed to a single return value using :func:`compose` decorator.

        :param message: The message to be executed (usually, a :class:`Command` or :class:`Query` subclass).
        :return: The result of the invoked message handler.

        :raises: ValueError: If no handlers are found for the message.
        """
        with self.transaction_context() as ctx:
            result = ctx.execute(message)
            return result

    def emit(self, event: Event) -> dict[Callable, Any]:
        """Deprecated. Use `publish()` instead."""
        with self.transaction_context() as ctx:
            result = ctx.emit(event)
        return result

    def on_enter_transaction_context(self, func):
        """
        Decorator for registering a function to be called when entering a transaction context

        :param func:
        :return:
        """
        self._on_enter_transaction_context = func
        return func

    def on_exit_transaction_context(self, func):
        """
        Decorator for registering a function to be called when exiting a transaction context

        :param func:
        :return:
        """
        self._on_exit_transaction_context = func
        return func

    def on_create_transaction_context(self, func):
        """
        Decorator for overriding default transaction context creation

        :param func:
        :return:
        """
        self._transaction_context_factory = func
        return func

    def transaction_middleware(self, middleware_func):
        """
        Decorator for registering a middleware function to be called when executing a function in a transaction context
        :param middleware_func:
        :return:
        """
        self._transaction_middlewares.insert(0, middleware_func)
        return middleware_func

    def compose(self, alias):
        """
        Decorator for composing results of handlers identified by an alias
        """

        def decorator(func):
            self._composers[alias] = func
            return func

        return decorator

    def transaction_context(self, **dependencies) -> TransactionContext:
        """Creates a transaction context for the app.

        The lifecycle of a transaction context is controlled by :func:`transaction_middleware`,
        :func:`on_enter_transaction_context`, :func:`on_exit_transaction_context` decorators.

        :param dependencies: Keyword arguments to be passed as dependencies to the dependency provider of a transaction
            context, will override dependencies provided by the application.
        :return: transaction context
        """
        if self._transaction_context_factory:
            ctx = self._transaction_context_factory(**dependencies)
        else:
            dp = self.dependency_provider.copy(**dependencies)
            ctx = TransactionContext(dependency_provider=dp)

        ctx.configure(
            on_enter_transaction_context=self._on_enter_transaction_context,
            on_exit_transaction_context=self._on_exit_transaction_context,
            middlewares=self._transaction_middlewares,
            composers=self._composers,
            handlers_iterator=self.iterate_handlers_for,
        )
        return ctx



import contextlib

from lato import Application


@contextlib.contextmanager
def override_app(application: Application, *args, **kwargs) -> Application:
    original_dependency_provider = application.dependency_provider
    overriden_dependency_provider = original_dependency_provider.copy(*args, **kwargs)

    application.dependency_provider = overriden_dependency_provider
    yield application

    application.dependency_provider = original_dependency_provider


@contextlib.contextmanager
def override_ctx(application: Application, *args, **kwargs) -> Application:
    original_transaction_context = application.transaction_context

    def overriden_transaction_context(**dependencies):
        ctx = original_transaction_context(**dependencies)
        ctx.dependency_provider = ctx.dependency_provider.copy(*args, **kwargs)
        return ctx

    application.transaction_context = overriden_transaction_context
    yield application

    application.transaction_context = original_transaction_context


@contextlib.contextmanager
def override(application: Application, *args, **kwargs) -> Application:
    with override_app(application, **kwargs) as overridden1:
        with override_ctx(overridden1, **kwargs) as overridden2:
            yield overridden2



import logging
import typing
from logging import NullHandler

from .application import Application
from .application_module import ApplicationModule
from .compositon import compose
from .dependency_provider import BasicDependencyProvider, DependencyProvider, as_type
from .message import Command, Event, Query
from .transaction_context import TransactionContext

__all__ = [
    "Application",
    "ApplicationModule",
    "DependencyProvider",
    "BasicDependencyProvider",
    "TransactionContext",
    "Command",
    "Query",
    "Event",
    "as_type",
    "compose",
]


logging.getLogger(__name__).addHandler(NullHandler())


def add_stderr_logger(
    level: int = logging.DEBUG,
) -> logging.StreamHandler[typing.TextIO]:
    """
    Helper for quickly adding a StreamHandler to the logger. Useful for
    debugging.

    Returns the handler after adding it.
    """
    # This method needs to be in this __init__.py to get the __name__ correct
    # even if lato is vendored within another package.
    logger = logging.getLogger(__name__)
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(message)s"))
    logger.addHandler(handler)
    logger.setLevel(level)
    logger.debug("Added a stderr logging handler to logger: %s", __name__)
    return handler


# ... Clean up.
del NullHandler



import logging
from collections import OrderedDict
from collections.abc import Callable, Iterator
from functools import partial
from typing import Any, NewType, Optional

from lato.compositon import compose
from lato.dependency_provider import (
    BasicDependencyProvider,
    DependencyProvider,
    as_type,
)
from lato.message import Message

Alias = NewType("Alias", Any)

log = logging.getLogger(__name__)


class TransactionContext:
    """Transaction context is a context manager for handler execution.

    :param dependency_provider_factory: a factory that returns :class:`DependencyProvider` instance,
        defaults to BasicDependencyProvider.

    **Example:**

    >>> from lato import TransactionContext
    >>>
    >>> def my_function(param1, param2):
    >>>     print(param1, param2)
    >>>
    >>> with TransactionContext(param1="foo") as ctx:
    >>>     ctx.call(my_function, param2="bar")
    foo, bar
    """

    dependency_provider_factory = BasicDependencyProvider

    def __init__(
        self, dependency_provider: DependencyProvider | None = None, *args, **kwargs
    ):
        """Initialize the transaction context instance.

        :param dependency_provider: dependency provider :class:`DependencyProvider` instance.
            Defaults to :class:`BasicDependencyProvider` instance populated with args and kwargs.
        :param args: Additional positional arguments to be passed to the dependency provider.
        :param kwargs: Additional keyword arguments to be passed to the dependency provider.
        """

        self.dependency_provider = (
            dependency_provider or self.dependency_provider_factory(*args, **kwargs)
        )
        self.resolved_kwargs: dict[str, Any] = {}
        self.current_handler: Optional[Callable] = None
        self._on_enter_transaction_context = lambda ctx: None
        self._on_exit_transaction_context = lambda ctx, exception=None: None
        self._middlewares: list[Callable] = []
        self._composers: dict[str | Message, Callable] = {}
        self._handlers_iterator: Iterator = lambda alias: iter([])

    def configure(
        self,
        on_enter_transaction_context=None,
        on_exit_transaction_context=None,
        middlewares=None,
        composers=None,
        handlers_iterator=None,
    ):
        """Customize the behavior of the transaction context with callbacks, middlewares, and composers.

        :param on_enter_transaction_context: Optional; Function to be called when entering a transaction context.
        :param on_exit_transaction_context: Optional; Function to be called when exiting a transaction context.
        :param middlewares: Optional; List of middleware functions to be applied.
        :param composers: Optional; List of composers functions to be applied.
        :param handlers_iterator: Optional; Function to iterate over handlers.
        """
        if on_enter_transaction_context:
            self._on_enter_transaction_context = on_enter_transaction_context
        if on_exit_transaction_context:
            self._on_exit_transaction_context = on_exit_transaction_context
        if middlewares:
            self._middlewares = middlewares
        if composers:
            self._composers = composers
        if handlers_iterator:
            self._handlers_iterator = handlers_iterator

    def begin(self):
        """Starts a transaction by calling `on_enter_transaction_context` callback.

        The callback could be used to set up the transaction-level dependencies (i.e. current time, transaction id),
        or to start the database transaction.
        """
        log.debug("Beginning transaction")
        """Should be used to start a transaction"""
        self._on_enter_transaction_context(self)

    def end(self, exception: Exception = None):
        """Ends the transaction context by calling `on_exit_transaction_context` callback,
        optionally passing an exception.

        The callback could be used to commit/end a database transaction.

        :param exception: Optional; The exception to handle at the end of the transaction, if any.
        """
        self._on_exit_transaction_context(self, exception)

        if exception:
            log.debug("Ended transaction with exception: {}".format(exception))
        else:
            log.debug("Ended transaction")

    def iterate_handlers_for(self, alias: str):
        yield from self._handlers_iterator(alias)

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        self.end(exc_val)

    def _wrap_with_middlewares(self, handler_func):
        p = handler_func
        for middleware in self._middlewares:
            p = partial(middleware, self, p)
        return p

    def call(self, func: Callable, *func_args: Any, **func_kwargs: Any) -> Any:
        """Call a function with the arguments and keyword arguments.
        Missing arguments will be resolved with the dependency provider.

        :param func: The function to call.
        :param func_args: Positional arguments to pass to the function.
        :param func_kwargs: Keyword arguments to pass to the function.
        :return: The result of the function call.
        """
        self.dependency_provider.update(ctx=as_type(self, TransactionContext))

        resolved_kwargs = self.dependency_provider.resolve_func_params(
            func, func_args, func_kwargs
        )
        self.resolved_kwargs.update(resolved_kwargs)
        p = partial(func, **resolved_kwargs)
        wrapped_handler = self._wrap_with_middlewares(p)
        result = wrapped_handler()
        return result

    def execute(self, message: Message) -> tuple[Any, ...]:
        """Executes all handlers bound to the message. Returns a tuple of handlers' return values.

        :param message: The message to be executed.
        :return: a tuple of return values from executed handlers
        :raises: ValueError: If no handlers are found for the message.
        """
        results = self.publish(message)
        values = tuple(results.values())

        if len(values) == 0:
            raise ValueError("No handlers found for message", values)

        composed_result = self._compose_results(message, values)
        return composed_result

    def emit(self, message: str | Message, *args, **kwargs) -> dict[Callable, Any]:
        # TODO: mark as obsolete
        return self.publish(message, *args, **kwargs)

    def publish(self, message: str | Message, *args, **kwargs) -> dict[Callable, Any]:
        """
        Publish a message by calling all handlers for that message.

        :param message: The message object to publish, or a string.
        :param args: Positional arguments to pass to the handlers.
        :param kwargs: Keyword arguments to pass to the handlers.
        :return: A dictionary mapping handlers to their results.
        """
        alias = type(message) if isinstance(message, Message) else message

        if isinstance(message, Message):
            args = (message, *args)

        all_results = OrderedDict()
        for handler in self._handlers_iterator(alias):
            self.set_dependency("message", message)
            # FIXME: push and pop current action instead of setting it
            self.current_handler = handler
            result = self.call(handler, *args, **kwargs)
            all_results[handler] = result
        return all_results

    def get_dependency(self, identifier: Any) -> Any:
        """Gets a dependency from the dependency provider"""
        return self.dependency_provider.get_dependency(identifier)

    def set_dependency(self, identifier: Any, dependency: Any) -> None:
        """Sets a dependency in the dependency provider"""
        self.dependency_provider.register_dependency(identifier, dependency)

    def set_dependencies(self, **kwargs):
        # TODO: add *args
        """Sets multiple dependencies at once"""
        self.dependency_provider.update(**kwargs)

    def __getitem__(self, item) -> Any:
        return self.get_dependency(item)

    def _compose_results(self, message: Message, results: tuple[Any, ...]) -> Any:
        alias = message.__class__  # TODO: expose alias as static field in Message class
        composer = self._composers.get(alias, compose)
        return composer(results)

    @property
    def current_action(self) -> tuple[Message, Callable]:
        """Returns current message and handler being executed"""
        return self.get_dependency("message"), self.current_handler
